"""vericlient module."""
from vericlient.daspeak.client import DaspeakClient
from vericlient.environments import Environments, Locations

__all__ = [
    "DaspeakClient",
    "Locations",
    "Environments",
]



"""Module to define the exceptions for the Daspeak API."""
from vericlient.exceptions import VeriClientError


class DaspeakError(VeriClientError):
    """Base class for exceptions in the Daspeak API."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class AudioInputError(DaspeakError):
    """Exception raised for errors in the audio input."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class TooManyAudioChannelsError(AudioInputError):
    """Exception raised for too many audio channels."""

    def __init__(self) -> None:
        message = "The maximum allowed number of audio channels is 2, and the audio provided has more channels"
        super().__init__(message)


class UnsupportedSampleRateError(AudioInputError):
    """Exception raised for unsupported sample rates."""

    def __init__(self) -> None:
        message = "The sample rate of the audio is not supported, must be 8 Khz or 16 Khz"
        super().__init__(message)


class AudioDurationTooLongError(AudioInputError):
    """Exception raised for audio duration too long."""

    def __init__(self) -> None:
        message = "The audio duration is too long, must be less than 30 seconds"
        super().__init__(message)


class UnsupportedAudioCodecError(AudioInputError):
    """Exception raised for unsupported audio codec."""

    def __init__(self) -> None:
        message = "The audio codec is not supported. Supported codecs are: 'PCM_16', 'ULAW', 'ALAW'"
        super().__init__(message)


class SignalNoiseRatioError(DaspeakError):
    """Exception raised for errors in the signal noise ratio."""

    def __init__(self) -> None:
        message = "Noise level of the audio exceeded"
        super().__init__(message)


class NetSpeechDurationIsNotEnoughError(DaspeakError):
    """Exception raised for errors in the net speech duration."""

    def __init__(self, net_speech_detected: float) -> None:
        message = (
            f"You need at least 3 seconds of speech to perform the operation, "
            f"but only {net_speech_detected} seconds were detected"
        )
        super().__init__(message)


class InvalidSpecifiedChannelError(DaspeakError):
    """Exception raised for invalid specified channel."""

    def __init__(self) -> None:
        message = "The specified channel is invalid, must be 1 or 2"
        super().__init__(message)


class InsufficientQualityError(DaspeakError):
    """Exception raised for insufficient quality."""

    def __init__(self) -> None:
        message = "The audio quality is insufficient or may contain more than one speaker"
        super().__init__(message)


class CalibrationNotAvailableError(DaspeakError):
    """Exception raised for calibration not available."""

    def __init__(self, calibration: str) -> None:
        message = f"The calibration {calibration} is not available"
        super().__init__(message)



"""Environments that the client can make requests to."""

from enum import Enum


class Environments(Enum):
    """Represents the environments that the client can make requests to, if the target is the cloud."""

    SANDBOX = "sandbox"
    PRODUCTION = "production"


class Locations(Enum):
    """Represents the locations that the client can make requests to, if the target is the cloud."""

    EU = "eu"
    US = "us"


cloud_env2url = {
    Environments.SANDBOX.value: {
        Locations.EU.value: f"https://api-work.{Locations.EU.value}.veri-das.com",
        Locations.US.value: f"https://api-work.{Locations.US.value}.veri-das.com",
    },
    Environments.PRODUCTION.value: {
        Locations.EU.value: f"https://api.{Locations.EU.value}.veri-das.com",
        Locations.US.value: f"https://api.{Locations.US.value}.veri-das.com",
    },
}



import pytest
import requests_mock

# pytest hooks

def pytest_addoption(parser):
    parser.addoption(
        "--mock", action="store_true", default=False, help="Run tests with mock server responses",
    )
    parser.addoption(
        "--env", action="store", default=None, help=\
            "Specify the environment to run tests against (e.g., EU_SANDBOX, US_PRODUCTION)",
    )


# variables

eu_sandbox_url = "https://api-work.eu.veri-das.com"
ue_production_url = "https://api.eu.veri-das.com"
us_sandbox_url = "https://api-work.us.veri-das.com"
us_production_url = "https://api.us.veri-das.com"

eu_sandobox_test_env = "EU_SANDBOX"
eu_production_test_env = "EU_PRODUCTION"
us_sandbox_test_env = "US_SANDBOX"
us_production_test_env = "US_PRODUCTION"


# general fixtures

@pytest.fixture(scope="session")
def mock_option(request):
    return request.config.getoption("--mock")


@pytest.fixture(scope="session")
def test_environment(request):
    return request.config.getoption("--env")


@pytest.fixture(scope="session")
def mock_server(mock_option, test_environment):
    if mock_option:
        with requests_mock.Mocker() as m:
            yield m
    else:
        if not test_environment:
            pytest.fail("No environment specified for real tests. Use --env to specify one.")
        yield None


@pytest.fixture(scope="session")
def all_environments():
    return [
        eu_sandobox_test_env,
        eu_production_test_env,
        us_sandbox_test_env,
        us_production_test_env,
    ]



"""General exceptions for the VeriClient package."""
from requests.models import Response


class VeriClientError(Exception):
    """Base class for exceptions in the VeriClient package."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class ServerError(VeriClientError):
    """Exception raised for server errors."""

    def __init__(self, response: Response) -> None:
        message = f"The server encountered an error with status code {response.status_code}: {response}"
        super().__init__(message)


class AuthorizationError(VeriClientError):
    """Exception raised for authorization errors."""

    def __init__(self) -> None:
        message = "The request was not authorized. Did you provide the correct API key?"
        super().__init__(message)


class UnsupportedMediaTypeError(VeriClientError):
    """Exception raised for unsupported media type errors."""

    def __init__(self) -> None:
        message = "The media type of the request is not supported."
        super().__init__(message)


class InvalidCredentialError(VeriClientError):
    """Exception raised for invalid credentials."""

    def __init__(self) -> None:
        message = "The credential/s provided are invalid."
        super().__init__(message)



"""Module to define the endpoints for Daspeak API."""
from enum import Enum

from vericlient.endpoints import Endpoints


class DaspeakEndpoints(Enum):   # noqa: D101
    ALIVE = Endpoints.ALIVE.value
    MODELS = "models"
    MODELS_METADATA = "models/metadata"
    MODELS_CALIBRATION = "models/calibration"
    MODELS_METADATA_FROM_CREDENTIAL = "models/metadata/from-credential"
    MODELS_HASH_CREDENTIAL_AUDIO = "models/<hash>/credential/wav"
    SIMILARITY_CREDENTIAL2CREDENTIAL = "similarity/credential2credential"
    SIMILARITY_CREDENTIAL2AUDIO = "similarity/credential2wav"
    SIMILARITY_AUDIO2AUDIO = "similarity/wav2wav"
    IDENTIFICATION_AUDIO2CREDENTIALS = "identification/wav2credentials"
    IDENTIFICATION_CREDENTIAL2CREDENTIALS = "identification/credential2credentials"



import pytest
from vericlient import DaspeakClient
from vericlient.daspeak.models import (
    CompareAudio2AudioInput,
    CompareAudio2AudioOutput,
    CompareAudio2CredentialsInput,
    CompareAudio2CredentialsOutput,
    CompareCredential2AudioInput,
    CompareCredential2AudioOutput,
    CompareCredential2CredentialInput,
    CompareCredential2CredentialOutput,
    CompareCredential2CredentialsInput,
    CompareCredential2CredentialsOutput,
    GenerateCredentialInput,
    GenerateCredentialOutput,
)


def test_daspeak_alive(mock_server, daspeak_alive_parameters):
    for param in daspeak_alive_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        if mock_server:
            mock_server.get(endpoint, json=mock_response, status_code=mock_status_code)

        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        response = daspeak_client.alive()

        assert response


def test_daspeak_get_models(mock_server, daspeak_get_models_parameters):
    for param in daspeak_get_models_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        if mock_server:
            mock_server.get(endpoint, json=mock_response, status_code=mock_status_code)

        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        response = daspeak_client.get_models()

        if mock_server:
            assert response.models == mock_response["models"]
        else:
            assert isinstance(response.models, list)


def test_daspeak_generate_credential(mock_server, daspeak_generate_credential_parameters, audio_file_path, audio_file):
    for param in daspeak_generate_credential_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
        else:
            model = daspeak_client.get_models().models[-1]

        input_model = GenerateCredentialInput(
            audio=audio_file_path,
            hash=model,
        )
        response = daspeak_client.generate_credential(input_model)
        assert isinstance(response, GenerateCredentialOutput)

        input_model = GenerateCredentialInput(
            audio=audio_file,
            hash=model,
        )
        response = daspeak_client.generate_credential(input_model)
        assert isinstance(response, GenerateCredentialOutput)


def _test_error(        # noqa: ANN202
    mock_server, daspeak_generate_credential_error_response_parameters, audio_file,
):
    for param in daspeak_generate_credential_error_response_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, exception = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
        else:
            model = daspeak_client.get_models().models[-1]

        input_model = GenerateCredentialInput(
            audio=audio_file,
            hash=model,
        )
        with pytest.raises(exception):
            daspeak_client.generate_credential(input_model)


def test_daspeak_generate_credential_too_many_audio_channels_error(
    mock_server, daspeak_generate_credential_channels_error_response_parameters, audio_too_many_channels_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_channels_error_response_parameters,
        audio_too_many_channels_file,
    )

def test_daspeak_generate_credential_sample_rate_error(
    mock_server, daspeak_generate_credential_sample_rate_error_response_parameters, audio_invalid_sample_rate_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_sample_rate_error_response_parameters,
        audio_invalid_sample_rate_file,
    )


def test_daspeak_generate_credential_net_speech_error(
    mock_server, daspeak_generate_credential_net_speech_duration_error_response_parameters, audio_not_enough_speech_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_net_speech_duration_error_response_parameters,
        audio_not_enough_speech_file,
    )


def test_daspeak_generate_credential_bad_snr_error(
    mock_server, daspeak_generate_credential_bad_snr_error_response_parameters, audio_bad_snr_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_bad_snr_error_response_parameters,
        audio_bad_snr_file,
    )


def test_daspeak_generate_credential_audio_too_long_error(
    mock_server, daspeak_generate_credential_audio_too_long_error_response_parameters, audio_too_long_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_audio_too_long_error_response_parameters,
        audio_too_long_file,
    )


def test_daspeak_generate_credentail_codec_error(
    mock_server, daspeak_generate_credential_codec_error_response_parameters, audio_codec_error_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_codec_error_response_parameters,
        audio_codec_error_file,
    )


def test_daspeak_generate_credential_insufficient_quality_error(
    mock_server, daspeak_generate_credential_insufficient_quality_error_response_parameters, audio_insufficient_quality_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_insufficient_quality_error_response_parameters,
        audio_insufficient_quality_file,
    )


def test_daspeak_generate_credential_invalid_specified_channel_error(
    mock_server, daspeak_generate_credential_invalid_specified_channel_error_response_parameters, audio_file,
):
    for param in daspeak_generate_credential_invalid_specified_channel_error_response_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, exception = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
        else:
            model = daspeak_client.get_models().models[-1]

        input_model = GenerateCredentialInput(
            audio=audio_file,
            hash=model,
            channel=100,
        )
        with pytest.raises(exception):
            daspeak_client.generate_credential(input_model)


def test_daspeak_generate_credential_calibration_not_available_error(
    mock_server, daspeak_generate_credential_calibration_not_available_error_response_parameters, audio_file,
):
    for param in daspeak_generate_credential_calibration_not_available_error_response_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, exception = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
        else:
            model = daspeak_client.get_models().models[-1]

        input_model = GenerateCredentialInput(
            audio=audio_file,
            hash=model,
            calibration="invalid-calibration",
        )
        with pytest.raises(exception):
            daspeak_client.generate_credential(input_model)


def test_daspeak_generate_credential_unsupported_media_type_error(
    mock_server, daspeak_generate_credential_unsupported_media_type_error_response_parameters, audio_unsupported_file,
):
    _test_error(
        mock_server,
        daspeak_generate_credential_unsupported_media_type_error_response_parameters,
        audio_unsupported_file,
    )


def test_daspeak_server_error(
    mock_server, daspeak_server_error_response_parameters, audio_file,
):
    for param in daspeak_server_error_response_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, exception = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"

            input_model = GenerateCredentialInput(
                audio=audio_file,
                hash=model,
            )
            with pytest.raises(exception):
                daspeak_client.generate_credential(input_model)


def test_daspeak_compare_credential2audio(
    mock_server, daspeak_compare_credential2audio_parameters, audio_file_path, audio_file,
):
    for param in daspeak_compare_credential2audio_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
            credential_reference = "fake-credential"
        else:
            model = daspeak_client.get_models().models[-1]
            credential_reference = daspeak_client.generate_credential(
                GenerateCredentialInput(audio=audio_file_path, hash=model),
            ).credential

        input_model = CompareCredential2AudioInput(
            audio_to_evaluate=audio_file_path,
            credential_reference=credential_reference,
        )
        response = daspeak_client.compare(input_model)
        assert isinstance(response, CompareCredential2AudioOutput)

        input_model = CompareCredential2AudioInput(
            audio_to_evaluate=audio_file,
            credential_reference=credential_reference,
        )
        response = daspeak_client.compare(input_model)
        assert isinstance(response, CompareCredential2AudioOutput)


def test_daspeak_compare_audio2audio(
    mock_server, daspeak_compare_audio2audio_parameters, audio_file,
):
    for param in daspeak_compare_audio2audio_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
        else:
            model = daspeak_client.get_models().models[-1]

        input_model = CompareAudio2AudioInput(
            audio_to_evaluate=audio_file,
            audio_reference=audio_file,
            hash=model,
        )
        response = daspeak_client.compare(input_model)
        assert isinstance(response, CompareAudio2AudioOutput)


def test_daspeak_compare_credential2credential(
    mock_server, daspeak_compare_credential2credential_parameters, audio_file,
):
    for param in daspeak_compare_credential2credential_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
            credential_reference = "fake-credential"
        else:
            model = daspeak_client.get_models().models[-1]
            credential_reference = daspeak_client.generate_credential(
                GenerateCredentialInput(audio=audio_file, hash=model),
            ).credential

        input_model = CompareCredential2CredentialInput(
            credential_to_evaluate=credential_reference,
            credential_reference=credential_reference,
        )
        response = daspeak_client.compare(input_model)
        assert isinstance(response, CompareCredential2CredentialOutput)


def test_daspeak_compare_audio2credentials(
    mock_server, daspeak_compare_audio2credentials_parameters, audio_file,
):
    for param in daspeak_compare_audio2credentials_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
            credential_list = [("id1", "fake-credential1"), ("id2", "fake-credential2")]
        else:
            model = daspeak_client.get_models().models[-1]
            credential = daspeak_client.generate_credential(GenerateCredentialInput(audio=audio_file, hash=model)).credential
            credential_list = [("id1", credential), ("id2", credential)]

        input_model = CompareAudio2CredentialsInput(
            audio_reference=audio_file,
            credential_list=credential_list,
        )
        response = daspeak_client.compare(input_model)
        assert isinstance(response, CompareAudio2CredentialsOutput)


def test_daspeak_client_compare_credential2credentials(
    mock_server, daspeak_compare_credential2credentials_parameters, audio_file,
):
    for param in daspeak_compare_credential2credentials_parameters:
        endpoint, mock_response, mock_status_code, url, environment, location, _ = param
        daspeak_client = DaspeakClient(
            apikey="fake-apikey",
            environment=environment,
            location=location,
            url=url,
        )
        if mock_server:
            mock_server.post(endpoint, json=mock_response, status_code=mock_status_code)
            model = "fake-model"
            credential = "fake-credential"
            credential_list = [("id1", "fake-credential1"), ("id2", "fake-credential2")]
        else:
            model = daspeak_client.get_models().models[-1]
            credential = daspeak_client.generate_credential(GenerateCredentialInput(audio=audio_file, hash=model)).credential
            credential_list = [("id1", credential), ("id2", credential)]

        response = daspeak_client.compare(CompareCredential2CredentialsInput(
            credential_reference=credential,
            credential_list=credential_list,
        ))
        assert isinstance(response, CompareCredential2CredentialsOutput)


def test_daspeak_client_compare_with_invalid_object_type():
    daspeak_client = DaspeakClient(apikey="fake-apikey")
    with pytest.raises(TypeError):
        daspeak_client.compare(data_model="invalid-object-type")


def test_daspeak_client_invalid_file_path():
    invalid_audio_file_path = "invalid-file-path"
    daspeak_client = DaspeakClient(apikey="fake-apikey")
    with pytest.raises(FileNotFoundError) as excinfo:
        daspeak_client.generate_credential(GenerateCredentialInput(audio="invalid-file-path", hash="fake-hash"))
    assert f"File {invalid_audio_file_path} not found" in str(excinfo.value)



"""Configuration for the application."""
import os

from dynaconf import Dynaconf

current_directory = os.path.dirname(os.path.realpath(__file__))
basepath = os.path.dirname(__file__)
config_path = os.path.join(basepath, "config.yml")
settings = Dynaconf(
    root_path=current_directory,
    settings_files=[config_path],
    envvar_prefix="VERICLIENT",
)



"""Module to define the models for the Daspeak API."""
# ruff: noqa: N805, D102, ANN201

from pydantic import BaseModel, field_validator


class DaspeakResponse(BaseModel):
    """Base class for the Daspeak API responses.

    Attributes:
        version: The version of the API
        status_code: The status code of the response

    """

    version: str
    status_code: int


class ModelsOutput(DaspeakResponse):
    """Output class for the get models endpoint.

    Attributes:
        models: The available models in the service

    """

    models: list


class GenerateCredentialInput(BaseModel):
    """Input class for the generate credential endpoint.

    Attributes:
        audio: The audio to generate the credential with.
            It can be a path to a file or a bytes object
            with the audio content
        hash: The hash of the biometrics model to use
        channel: The `nchannel` of the audio if it is stereo
        calibration: The calibration to use

    """

    audio: str | bytes
    hash: str
    channel: int = 1
    calibration: str = "telephone-channel"

    @field_validator("audio")
    def must_be_str_or_bytes(cls, value: object):
        if not isinstance(value, (str, bytes)):
            error = "audio must be a string or a bytes object"
            raise TypeError(error)
        return value

    class Config:
        arbitrary_types_allowed = True


class ModelMetadata(BaseModel):
    """Metadata of the model used to generate the credential.

    Attributes:
        hash: The hash of the model
        mode: The mode of the model

    """

    hash: str
    mode: str


class GenerateCredentialOutput(DaspeakResponse):
    """Output class for the generate credential endpoint.

    Attributes:
        model: The model used to generate the credential
        credential: The generated credential
        authenticity: The authenticity of the audio sample used
        input_audio_duration: The duration of the input audio
        net_speech_duration: The duration of the speech in the audio

    """

    model: ModelMetadata
    credential: str
    authenticity: float
    input_audio_duration: float
    net_speech_duration: float

    @field_validator("authenticity")
    def round_authenticity(cls, value: float) -> float:
        return round(value, 3)


class CompareInput(BaseModel):
    """Base class for the similarity inputs.

    Attributes:
        calibration: The calibration to use

    """

    calibration: str = "telephone-channel"


class CompareOutput(DaspeakResponse):
    """Base class for the similarity outputs.

    Attributes:
        score: The similarity score between the two inputs

    """

    score: float

    @field_validator("score")
    def round_value(cls, value: float) -> float:
        return round(value, 3)


class CompareCredential2CredentialInput(CompareInput):
    """Input class for the similarity credential to credential endpoint.

    Attributes:
        credential_reference: The reference credential
        credential_to_evaluate: The credential to evaluate
        calibration: The calibration to use

    """

    credential_reference: str
    credential_to_evaluate: str


class CompareCredential2CredentialOutput(CompareOutput):
    """Output class for the similarity credential to credential endpoint.

    Attributes:
        calibration: The calibration used

    """

    calibration: str


class CompareCredential2AudioInput(CompareInput):
    """Input class for the similarity credential to audio endpoint.

    Attributes:
        credential_reference: The reference credential
        audio_to_evaluate: The audio to evaluate.
            It can be a path to a file or a bytes object
            with the audio content
        channel: The `nchannel` of the audio if it is stereo

    """

    credential_reference: str
    audio_to_evaluate: str | bytes
    channel: int = 1

    @field_validator("audio_to_evaluate")
    def must_be_str_or_bytes(cls, value: object):
        if not isinstance(value, (str, bytes)):
            error = "audio must be a string or a bytes object"
            raise TypeError(error)
        return value

    class Config:
        arbitrary_types_allowed = True


class CompareCredential2AudioOutput(CompareOutput):
    """Output class for the similarity credential to audio endpoint.

    Attributes:
        model: The model used to generate the credential
        calibration: The calibration used
        authenticity_to_evaluate: The authenticity of the audio sample used
        input_audio_duration: The duration of the input audio
        net_speech_duration: The duration of the speech in the audio

    """

    model: ModelMetadata
    calibration: str
    authenticity_to_evaluate: float
    input_audio_duration_to_evaluate: float
    net_speech_duration_to_evaluate: float

    @field_validator("authenticity_to_evaluate")
    def round_value(cls, value: float) -> float:
        return round(value, 3)


class CompareAudio2AudioInput(CompareInput):
    """Input class for the similarity audio to audio endpoint.

    Attributes:
        audio_reference: The reference audio.
            It can be a path to a file or a bytes object
            with the audio content
        audio_to_evaluate: The audio to evaluate.
            It can be a path to a file or a bytes object
            with the audio content
        channel_reference: The `nchannel` of the reference audio if it is stereo
        channel_to_evaluate: The `nchannel` of the audio to evaluate if it is stereo

    """

    audio_reference: str | bytes
    audio_to_evaluate: str | bytes
    channel_reference: int = 1
    channel_to_evaluate: int = 1

    @field_validator("audio_reference", "audio_to_evaluate")
    def audio_ref_must_be_str_or_bytes(cls, value: object):
        if not isinstance(value, (str, bytes)):
            error = "audio must be a string or a bytes object"
            raise TypeError(error)
        return value

    class Config:
        arbitrary_types_allowed = True


class CompareAudio2AudioOutput(CompareOutput):
    """Output class for the similarity audio to audio endpoint.

    Attributes:
        model: The model used to generate the credential
        calibration: The calibration used
        authenticity_reference: The authenticity of the reference audio sample
        authenticity_to_evaluate: The authenticity of the audio to evaluate
        input_audio_duration_reference: The duration of the reference audio
        input_audio_duration_to_evaluate: The duration of the audio to evaluate
        net_speech_duration_reference: The duration of the speech in the reference audio
        net_speech_duration_to_evaluate: The duration of the speech in the audio to evaluate

    """

    model: ModelMetadata
    calibration: str
    authenticity_reference: float
    authenticity_to_evaluate: float
    input_audio_duration_reference: float
    input_audio_duration_to_evaluate: float
    net_speech_duration_reference: float
    net_speech_duration_to_evaluate: float

    @field_validator("authenticity_reference", "authenticity_to_evaluate")
    def round_value(cls, value: float) -> float:
        return round(value, 3)


class CompareAudio2CredentialsInput(CompareInput):
    """Input class for the identification audio to credentials endpoint.

    Attributes:
        audio_reference: The audio to evaluate.
            It can be a path to a file or a bytes object
            with the audio content
        credential_list: The credentials to compare the audio with.
            The list contains touples with two strings: the id and the credential
        channel: The `nchannel` of the audio if it is stereo

    """

    audio_reference: str | bytes
    credential_list: list[tuple[str, str]]
    channel: int = 1

    @field_validator("audio_reference")
    def must_be_str_or_bytes(cls, value: object):
        if not isinstance(value, (str, bytes)):
            error = "audio must be a string or a bytes object"
            raise TypeError(error)
        return value

    @field_validator("credential_list")
    def validate_and_build_list_format(cls, value: list):
        if not value:
            error = "credential_list must not be empty"
            raise ValueError(error)
        error = "credential_list must contain touples with two strings"
        n_items = 2
        for item in value:
            if not isinstance(item, tuple) or len(item) != n_items:
                raise ValueError(error)
            if not all(isinstance(i, str) for i in item):
                raise ValueError(error)
        return [{"id": item[0], "credential": item[1]} for item in value]

    class Config:
        arbitrary_types_allowed = True


class CompareAudio2CredentialsOutput(DaspeakResponse):
    """Output class for the identification audio to credentials endpoint.

    Attributes:
        result: The result of the identification, a dictionary with the "id" and the "score"
            of the best match
        scores: The whole list of scores for each credential.
            The list contains dictionaries with two keys (and values): "id" and "score"
        calibration: The calibration used
        model: The model used to generate the credential
        authenticity_reference: The authenticity of the reference audio sample
        input_audio_duration_reference: The duration of the input audio
        net_speech_duration_reference: The duration of the speech in the audio

    """

    result: dict
    scores: list[dict]
    calibration: str
    model: ModelMetadata
    authenticity_reference: float
    input_audio_duration_reference: float
    net_speech_duration_reference: float

    @field_validator("authenticity_reference")
    def round_value(cls, value: float) -> float:
        return round(value, 3)


class CompareCredential2CredentialsInput(CompareInput):
    """Input class for the identification credential to credentials endpoint.

    Attributes:
        credential_reference: The reference credential
        credential_list: The credentials to compare the audio with.
            The list contains touples with two strings: the id and the credential

    """

    credential_reference: str
    credential_list: list[tuple[str, str]]

    @field_validator("credential_list")
    def validate_and_build_list_format(cls, value: list):
        if not value:
            error = "credential_list must not be empty"
            raise ValueError(error)
        error = "credential_list must contain touples with two strings"
        n_items = 2
        for item in value:
            if not isinstance(item, tuple) or len(item) != n_items:
                raise ValueError(error)
            if not all(isinstance(i, str) for i in item):
                raise ValueError(error)
        return [{"id": item[0], "credential": item[1]} for item in value]


class CompareCredential2CredentialsOutput(DaspeakResponse):
    """Output class for the identification credential to credentials endpoint.

    Attributes:
        result: The result of the identification, a dictionary with the "id" and the "score"
            of the best match
        scores: The whole list of scores for each credential.
            The list contains dictionaries with two keys (and values): "id" and "score"
        calibration: The calibration used
        model: The model used to generate the credential

    """

    result: dict
    scores: list[dict]
    calibration: str



"""Module with the abstraction of the client to interact with the Veridas APIs."""
from abc import ABC, abstractmethod

import requests
import structlog

from vericlient.apis import APIs
from vericlient.config.config import settings
from vericlient.environments import Environments, Locations, cloud_env2url
from vericlient.exceptions import AuthorizationError, ServerError

logger = structlog.get_logger(__name__)


class Client(ABC):
    """Class to interact with the Veridas APIs."""

    def __init__(
            self,
            api: str,
            apikey: str | None = None,
            timeout: int | None = None,
            environment: str | None = None,
            location: str | None = None,
            url: str | None = None,
            headers: dict | None = None,
    ) -> None:
        """Create Client class."""
        self._headers = headers or {}
        self._session = requests.Session()

        if not timeout and not settings.timeout:
            seconds = 10
            logging_message = f"No timeout provided. Defaulting to {seconds} seconds"
            logger.warning(logging_message)
            self._timeout = seconds
        else:
            self._timeout = settings.timeout or timeout

        if url:
            self._configure_custom_url(url)
        else:
            self._configure_cloud_url(api, environment, location)
            if not apikey and not settings.apikey:
                error = "If target is cloud, apikey must be provided"
                raise ValueError(error)
            apikey = settings.apikey or apikey
            self._headers.update({"apikey": apikey})

        self._session.headers.update(self._headers)

    def _configure_cloud_url(self, api: str, environment: str, location: str) -> None:
        if not environment and not settings.environment:
            logger.warning("No environment provided. Defaulting to sandbox")
            environment = Environments.SANDBOX.value
        else:
            environment = settings.environment or environment
        if not any(environment == env.value for env in Environments):
            error = f"Invalid environment: {environment}. Valid options are: {', '.join(env.value for env in Environments)}"
            raise ValueError(error)

        if not location and not settings.location:
            logger.warning("No location provided. Defaulting to EU")
            location = Locations.EU.value
        else:
            location = settings.location or location
        if not any(location == loc.value for loc in Locations):
            error = f"Invalid location: {location}. Valid options are: {', '.join(loc.value for loc in Locations)}"
            raise ValueError(error)

        if not any(api == api_.value for api_ in APIs):
            error = f"If target is cloud, valid api must be provided. Valid options are: {', '.join(api.value for api in APIs)}"
            raise ValueError(error)
        self._url = cloud_env2url[environment][location] + f"/{api}"

    def _configure_custom_url(self, url: str) -> None:
        url = settings.url or url
        self._url = url

    @property
    def url(self) -> str:
        """Return the URL of the API."""
        return self._url

    @property
    def headers(self) -> dict:
        """Return the headers of the API."""
        return self._headers

    @property
    def timeout(self) -> int:
        """Return the timeout of the API."""
        return self._timeout

    @abstractmethod
    def alive(self) -> bool:
        """Check if the API is alive and responding."""

    @abstractmethod
    def _handle_error_response(self, response: requests.Response) -> None:
        """Handle error responses from the API."""

    def _get(self, endpoint: str) -> requests.Response:
        """Make a GET request to the API."""
        response = self._session.get(f"{self._url}/{endpoint}", timeout=self._timeout)
        if not response.ok:
            self._handle_authorization_error(response)
            self._handle_error_response(response)
        return response

    def _post(
            self, endpoint: str,
            data: dict | None = None,
            json_: dict | None = None,
            files: dict | None = None,
    ) -> requests.Response:
        """Make a POST request to the API."""
        response = self._session.post(
            f"{self._url}/{endpoint}",
            data=data,
            json=json_,
            files=files,
            timeout=self._timeout,
        )
        if not response.ok:
            self._handle_authorization_error(response)
            self._handle_error_response(response)
        return response

    def _raise_server_error(self, response: requests.Response) -> None:
        """Raise a ServerError exception."""
        raise ServerError(response)

    def _handle_authorization_error(self, response: requests.Response) -> None:
        """Handle authorization errors."""
        try:
            message = response.json()["message"]
            if "no Authorization header found" in message:
                raise AuthorizationError
        except KeyError:
            pass



import pytest
from vericlient.daspeak.exceptions import (
    AudioDurationTooLongError,
    CalibrationNotAvailableError,
    InsufficientQualityError,
    InvalidSpecifiedChannelError,
    NetSpeechDurationIsNotEnoughError,
    SignalNoiseRatioError,
    TooManyAudioChannelsError,
    UnsupportedAudioCodecError,
    UnsupportedSampleRateError,
)
from vericlient.environments import Environments, Locations
from vericlient.exceptions import (
    ServerError,
    UnsupportedMediaTypeError,
)

from tests.conftest import (
    eu_production_test_env,
    eu_sandbox_url,
    eu_sandobox_test_env,
    ue_production_url,
    us_production_test_env,
    us_production_url,
    us_sandbox_test_env,
    us_sandbox_url,
)

####################
# SERVER RESPONSES #
####################

@pytest.fixture(scope="session")
def daspeak_alive_response():
    return ""


@pytest.fixture(scope="session")
def daspeak_get_models_response():
    return {"version": "1", "models": ["model1", "model2"]}


@pytest.fixture(scope="session")
def daspeak_generate_credential_response():
    return {
        "version": "1",
        "model": {
            "hash": "fake-hash",
            "mode": "fake-mode",
        },
        "credential": "fake-credential",
        "authenticity": 0.99,
        "input_audio_duration": 5.00,
        "net_speech_duration": 4.50,
    }


@pytest.fixture(scope="session")
def daspeak_compare_credential2audio_response():
    return {
        "version": "1",
        "score": 0.99,
        "model": {
            "hash": "fake-hash",
            "mode": "fake-mode",
        },
        "calibration": "fake-calibration",
        "authenticity_to_evaluate": 0.99,
        "input_audio_duration_to_evaluate": 5.00,
        "net_speech_duration_to_evaluate": 4.50,
    }


@pytest.fixture(scope="session")
def daspeak_compare_audio2audio_response():
    return {
        "version": "1",
        "score": 0.99,
        "model": {
            "hash": "fake-hash",
            "mode": "fake-mode",
        },
        "calibration": "fake-calibration",
        "authenticity_to_evaluate": 0.99,
        "authenticity_reference": 0.99,
        "input_audio_duration_to_evaluate": 5.00,
        "input_audio_duration_reference": 5.00,
        "net_speech_duration_to_evaluate": 4.50,
        "net_speech_duration_reference": 4.50,
    }


@pytest.fixture(scope="session")
def daspeak_compare_credential2credential_response():
    return {
        "version": "1",
        "score": 0.99,
        "model": {
            "hash": "fake-hash",
            "mode": "fake-mode",
        },
        "calibration": "fake-calibration",
    }


@pytest.fixture(scope="session")
def daspeak_compare_audio2credentials_response():
    return {
        "version": "1",
        "model": {
            "hash": "fake-hash",
            "mode": "fake-mode",
        },
        "calibration": "fake-calibration",
        "authenticity_reference": 0.99,
        "scores": [
            {
                "id": "fake-id1",
                "score": 0.99,
            },
            {
                "id": "fake-id2",
                "score": 0.97,
            },
        ],
        "result": {
            "id": "fake-id1",
            "score": 0.99,
        },
        "input_audio_duration_reference": 5.00,
        "net_speech_duration_reference": 4.50,
    }


@pytest.fixture(scope="session")
def daspeak_compare_credential2credentials_response():
    return {
        "version": "1",
        "calibration": "fake-calibration",
        "scores": [
            {
                "id": "fake-id1",
                "score": 0.99,
            },
            {
                "id": "fake-id2",
                "score": 0.97,
            },
        ],
        "result": {
            "id": "fake-id1",
            "score": 0.99,
        },
    }


#################
# SERVER ERRORS #
#################

@pytest.fixture(scope="session")
def daspeak_channels_error_response():
    return {
        "error": "The wav has more channels than are accepted by the system",
        "exception": "AudioInputException",
    }


@pytest.fixture(scope="session")
def daspeak_sample_rate_error_response():
    return {
        "error": "The sample rate is not supported, must be 8 Khz or 16 Khz",
        "exception": "AudioInputException",
    }


@pytest.fixture(scope="session")
def daspeak_codec_error_response():
    return {
        "error": "Provided audio data uses an unsupported codec. Supported codecs are:",
        "exception": "AudioInputException",
    }


@pytest.fixture(scope="session")
def daspeak_long_audio_error_response():
    return {
        "error": "The wav duration is longer than 30 seconds",
        "exception": "AudioInputException",
    }


@pytest.fixture(scope="session")
def daspeak_snr_error_response():
    return {
        "error": "Noise level exceeded",
        "exception": "SignalNoiseRatioException",
    }


@pytest.fixture(scope="session")
def daspeak_net_speech_duration_error_response():
    return {
        "error": "Voice duration is not enough 2.55s < 3.00s",
        "exception": "VoiceDurationIsNotEnoughException",
    }


@pytest.fixture(scope="session")
def daspeak_invalid_specified_channel_error_response():
    return {
        "error": "Invalid specified channel/s",
        "exception": "InvalidChannelException",
    }


@pytest.fixture(scope="session")
def daspeak_quality_error_response():
    return {
        "error": "The audio quality is not good enough",
        "exception": "InsufficientQuality",
    }


@pytest.fixture(scope="session")
def daspeak_calibration_not_available_error_response():
    return {
        "error": "The calibration <calibration> is not available",
        "exception": "CalibrationNotAvailable",
    }


@pytest.fixture(scope="session")
def daspeak_invalid_credential_error_response():
    return {
        "error": "Decryption error",
        "exception": "InvalidCredential",
    }


@pytest.fixture(scope="session")
def daspeak_unsupported_media_type_error_response():
    return {
        "error": "415 Unsupported Media Type: The server does not support the media type transmitted in the request",
        "exception": "UnsupportedMediaType",
    }


@pytest.fixture(scope="session")
def daspeak_server_error_response():
    return {
        "error": "An internal server error occured",
        "exception": "DaspeakInternalException",
    }


#######################
# PARAMETERS FIXTURES #
#######################

def _provide_daspeak_parameters(
        test_environment: str,
        all_environments: list,
        mock_option: bool,      # noqa: FBT001
        endpoint: str,
        response: dict,
        status_code: int,
        exception: Exception,
    ) -> list:
    """Provide the parameters necessary for Daspekea testing depending on the test environment.

    Those parameters are:
    - endpoint: the endpoint to test
    - response: the response to return
    - status_code: the status code to return
    - exception: the exception to raise if any

    The parameters are returned as a list of tuples, each tuple containing the parameters for a specific environment.

    """
    ue_sandbox = (f"{eu_sandbox_url}/{endpoint}", response, status_code, None, \
        Environments.SANDBOX.value, Locations.EU.value, exception)
    ue_production = (f"{ue_production_url}/{endpoint}", response, status_code, None, \
        Environments.PRODUCTION.value, Locations.EU.value, exception)
    us_sandbox = (f"{us_sandbox_url}/{endpoint}", response, status_code, None, \
        Environments.SANDBOX.value, Locations.US.value, exception)
    us_production = (f"{us_production_url}/{endpoint}", response, status_code, None, \
        Environments.PRODUCTION.value, Locations.US.value, exception)
    if mock_option:
        parameters = [
            ue_sandbox,
            ue_production,
            us_sandbox,
            us_production,
            (f"https://custom-daspeak-url.com/{endpoint}", response, status_code, \
                "https://custom-daspeak-url.com/daspeak/v1", None, None, exception),
        ]
    elif test_environment not in all_environments:
        pytest.fail(f"Invalid environment specified. Use one of {all_environments}")
    elif test_environment == eu_sandobox_test_env:
        parameters = [ue_sandbox]
    elif test_environment == eu_production_test_env:
        parameters = [ue_production]
    elif test_environment == us_sandbox_test_env:
        parameters = [us_sandbox]
    elif test_environment == us_production_test_env:
        parameters = [us_production]
    return parameters


@pytest.fixture(scope="session")
def daspeak_alive_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_alive_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/alive", \
        daspeak_alive_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_get_models_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_get_models_response,
    )-> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models", \
        daspeak_get_models_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_generate_credential_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_generate_credential_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_channels_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_channels_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_channels_error_response, 400, TooManyAudioChannelsError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_sample_rate_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_sample_rate_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_sample_rate_error_response, 400, UnsupportedSampleRateError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_net_speech_duration_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_net_speech_duration_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_net_speech_duration_error_response, 400, NetSpeechDurationIsNotEnoughError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_bad_snr_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_snr_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_snr_error_response, 400, SignalNoiseRatioError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_audio_too_long_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_long_audio_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_long_audio_error_response, 400, AudioDurationTooLongError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_codec_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_codec_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_codec_error_response, 400, UnsupportedAudioCodecError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_insufficient_quality_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_quality_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_quality_error_response, 400, InsufficientQualityError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_invalid_specified_channel_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_invalid_specified_channel_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_invalid_specified_channel_error_response, 400, InvalidSpecifiedChannelError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_calibration_not_available_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_calibration_not_available_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_calibration_not_available_error_response, 400, CalibrationNotAvailableError,
    )


@pytest.fixture(scope="session")
def daspeak_generate_credential_unsupported_media_type_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_unsupported_media_type_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_unsupported_media_type_error_response, 415, UnsupportedMediaTypeError,
    )


@pytest.fixture(scope="session")
def daspeak_server_error_response_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_server_error_response,
    ):
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/models/fake-model/credential/wav", \
        daspeak_server_error_response, 500, ServerError,
    )


@pytest.fixture(scope="session")
def daspeak_compare_credential2audio_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_compare_credential2audio_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/similarity/credential2wav", \
        daspeak_compare_credential2audio_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_compare_audio2audio_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_compare_audio2audio_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/similarity/wav2wav", \
        daspeak_compare_audio2audio_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_compare_credential2credential_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_compare_credential2credential_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/similarity/credential2credential", \
        daspeak_compare_credential2credential_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_compare_audio2credentials_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_compare_audio2credentials_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/identification/wav2credentials", \
        daspeak_compare_audio2credentials_response, 200, None,
    )


@pytest.fixture(scope="session")
def daspeak_compare_credential2credentials_parameters(
        mock_option,
        test_environment,
        all_environments,
        daspeak_compare_credential2credentials_response,
    ) -> list:
    return _provide_daspeak_parameters(
        test_environment, all_environments, mock_option, "daspeak/v1/identification/credential2credentials", \
        daspeak_compare_credential2credentials_response, 200, None,
    )


######################
# RESOURCES FIXTURES #
######################

@pytest.fixture(scope="session")
def audio_file_path() -> str:
    return "tests/daspeak/resources/audio.wav"


@pytest.fixture(scope="session")
def audio_file() -> bytes:
    with open("tests/daspeak/resources/audio.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_spoof_file() -> bytes:
    with open("tests/daspeak/resources/audio_spoof.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_not_enough_speech_file() -> bytes:
    with open("tests/daspeak/resources/audio_not_enough_speech.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_bad_snr_file() -> bytes:
    with open("tests/daspeak/resources/audio_bad_snr.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_insufficient_quality_file() -> bytes:
    with open("tests/daspeak/resources/audio_insufficient_quality.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_too_many_channels_file() -> bytes:
    with open("tests/daspeak/resources/audio_too_many_channels.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_invalid_sample_rate_file() -> bytes:
    with open("tests/daspeak/resources/audio_invalid_sample_rate.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_too_long_file() -> bytes:
    with open("tests/daspeak/resources/audio_too_long.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_codec_error_file() -> bytes:
    with open("tests/daspeak/resources/audio_codec_error.wav", "rb") as f:
        return f.read()


@pytest.fixture(scope="session")
def audio_unsupported_file() -> bytes:
    with open("tests/__init__.py", "rb") as f:
        return f.read()



"""Implementation of the client for the DASPEaK service."""
import json

from requests.models import Response

from vericlient.apis import APIs
from vericlient.client import Client
from vericlient.daspeak.endpoints import DaspeakEndpoints
from vericlient.daspeak.exceptions import (
    AudioDurationTooLongError,
    CalibrationNotAvailableError,
    InsufficientQualityError,
    InvalidSpecifiedChannelError,
    NetSpeechDurationIsNotEnoughError,
    SignalNoiseRatioError,
    TooManyAudioChannelsError,
    UnsupportedAudioCodecError,
    UnsupportedSampleRateError,
    VeriClientError,
)
from vericlient.daspeak.models import (
    CompareAudio2AudioInput,
    CompareAudio2AudioOutput,
    CompareAudio2CredentialsInput,
    CompareAudio2CredentialsOutput,
    CompareCredential2AudioInput,
    CompareCredential2AudioOutput,
    CompareCredential2CredentialInput,
    CompareCredential2CredentialOutput,
    CompareCredential2CredentialsInput,
    CompareCredential2CredentialsOutput,
    CompareInput,
    GenerateCredentialInput,
    GenerateCredentialOutput,
    ModelsOutput,
)
from vericlient.exceptions import InvalidCredentialError, UnsupportedMediaTypeError


class DaspeakClient(Client):
    """Class to interact with the Daspeak API."""

    def __init__(
            self,
            apikey: str | None = None,
            timeout: int | None = None,
            environment: str | None = None,
            location: str | None = None,
            url: str | None = None,
            headers: dict | None = None,
    ) -> None:
        """Create the DaspeakClient class.

        Args:
            apikey: The API key to use
            timeout: The timeout to use in the requests
            environment: The environment to use
            location: The location to use
            url: The URL to use in case of a custom target
            headers: The headers to be used in the requests

        """
        api = APIs.DASPEAK.value
        super().__init__(
            api=api,
            apikey=apikey,
            timeout=timeout,
            environment=environment,
            location=location,
            url=url,
            headers=headers,
        )
        self._exceptions = [
            "AudioInputException",
            "SignalNoiseRatioException",
            "VoiceDurationIsNotEnoughException",
            "InvalidChannelException",
            "InsufficientQuality",
            "CalibrationNotAvailable",
            "ServerError",
            "InvalidCredential",
            "UnsupportedMediaType",
        ]
        self._compare_functions_map = {
            CompareCredential2AudioInput: self._compare_credential2audio,
            CompareAudio2AudioInput: self._compare_audio2audio,
            CompareCredential2CredentialInput: self._compare_credential2credential,
            CompareAudio2CredentialsInput: self._compare_audio2credentials,
            CompareCredential2CredentialsInput: self._compare_credential2credentials,
        }
        self._exception_map = {
            "SignalNoiseRatioException": SignalNoiseRatioError,
            "VoiceDurationIsNotEnoughException": self._handle_voice_duration_error,
            "InvalidChannelException": InvalidSpecifiedChannelError,
            "InsufficientQuality": InsufficientQualityError,
            "CalibrationNotAvailable": self._handle_calibration_error,
            "InvalidCredential": InvalidCredentialError,
            "UnsupportedMediaType": UnsupportedMediaTypeError,
        }
        self._audio_input_errors = {
            "more channels than": TooManyAudioChannelsError,
            "unsupported codec": UnsupportedAudioCodecError,
            "sample rate": UnsupportedSampleRateError,
            "duration is longer": AudioDurationTooLongError,
        }

    def alive(self) -> bool:
        """Check if the service is alive.

        Returns
            bool: True if the service is alive, False otherwise

        """
        response = self._get(endpoint=DaspeakEndpoints.ALIVE.value)
        accepted_status_code = 200
        return response.status_code == accepted_status_code

    def _handle_error_response(self, response: Response) -> None:
        """Handle error responses from the API."""
        response_json = response.json()

        exception = response_json.get("exception")
        if not exception or exception not in self._exceptions:
            self._raise_server_error(response)

        if exception == "AudioInputException":
            error_message = response_json.get("error", "")
            for error_text, error_class in self._audio_input_errors.items():
                if error_text in error_message:
                    raise error_class
            raise ValueError(error_message)

        handler = self._exception_map.get(exception)
        if handler:
            if isinstance(handler, type) and issubclass(handler, VeriClientError):
                raise handler
            handler(response_json)

        raise ValueError(response_json.get("error", "Unknown error"))

    def _handle_voice_duration_error(self, response_json: dict) -> None:
        error_message = response_json.get("error", "")
        net_speech_detected = float(error_message.split(" ")[-3].replace("s", ""))
        raise NetSpeechDurationIsNotEnoughError(net_speech_detected)

    def _handle_calibration_error(self, response_json: dict) -> None:
        error_message = response_json.get("error", "")
        calibration = str(error_message.split(" ")[2])
        raise CalibrationNotAvailableError(calibration)

    def get_models(self) -> ModelsOutput:
        """Get the models available biometrics models in the service.

        Returns:
            The response from the service

        """
        response = self._get(endpoint=DaspeakEndpoints.MODELS.value)
        return ModelsOutput(status_code=response.status_code, **response.json())

    def generate_credential(self, data_model: GenerateCredentialInput) -> GenerateCredentialOutput:
        """Generate a credential from a WAV file.

        Args:
            data_model: The data required to generate the credential

        Returns:
            The response from the service

        Raises:
            ValueError: If the `data_model` is not an instance of `GenerateCredentialInput`
            TooManyAudioChannelsError: If the audio has more channels than the service supports
            UnsupportedAudioCodecError: If the audio has an unsupported codec
            UnsupportedSampleRateError: If the audio has an unsupported sample rate
            AudioDurationTooLongError: If the audio duration is longer than the service supports
            SignalNoiseRatioError: If the signal-to-noise ratio is too low
            NetSpeechDurationIsNotEnoughError: If the net speech duration is not enough
            InvalidSpecifiedChannelError: If the specified channel is invalid
            InsufficientQualityError: If the audio quality is insufficient
            CalibrationNotAvailableError: If the calibration is not available
            UnsupportedMediaTypeError: If the media type is not supported

        """
        endpoint = DaspeakEndpoints.MODELS_HASH_CREDENTIAL_AUDIO.value.replace("<hash>", data_model.hash)
        audio = self._get_virtual_audio_file(data_model.audio)
        files = {
            "audio": ("audio", audio, "audio/wav"),
        }
        data = {
            "channel": data_model.channel,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data, files=files)
        return GenerateCredentialOutput(status_code=response.status_code, **response.json())

    def compare(    # noqa: D417
            self,
            data_model: CompareInput,
        ) -> CompareCredential2AudioOutput | CompareAudio2AudioOutput | \
             CompareCredential2CredentialOutput | CompareAudio2CredentialsOutput | \
             CompareCredential2CredentialsOutput:
        """Compare two sets of data based on the provided input.

        Args:
            data_model (CompareCredential2AudioInput | CompareAudio2AudioInput | CompareCredential2CredentialInput | \
                        CompareAudio2CredentialsInput | CompareCredential2CredentialsInput):
                The data required to compare the audio files or credentials

        Returns:
            The response from the service, depending on the input type.

        Raises:
            ValueError: If the `data_model` is not an instance of `CompareInput`
            TooManyAudioChannelsError: If the audio has more channels than the service supports
            UnsupportedAudioCodecError: If the audio has an unsupported codec
            UnsupportedSampleRateError: If the audio has an unsupported sample rate
            AudioDurationTooLongError: If the audio duration is longer than the service supports
            SignalNoiseRatioError: If the signal-to-noise ratio is too low
            NetSpeechDurationIsNotEnoughError: If the net speech duration is not enough
            InvalidSpecifiedChannelError: If the specified channel is invalid
            InsufficientQualityError: If the audio quality is insufficient
            CalibrationNotAvailableError: If the calibration is not available
            InvalidCredentialError: If the credential is invalid
            UnsupportedMediaTypeError: If the media type is not supported

        """
        try:
            func = self._compare_functions_map.get(type(data_model))
            return func(data_model)
        except AttributeError as e:
            error = "data_model must be an instance of CompareInput"
            raise TypeError(error) from e

    def _compare_credential2audio(
            self,
            data_model: CompareCredential2AudioInput,
        ) -> CompareCredential2AudioOutput:
        """Compare a credential with an audio file.

        Args:
            data_model: The data required to compare the credential with the audio

        Returns:
            CompareCredential2AudioOutput: The response from the service

        """
        endpoint = DaspeakEndpoints.SIMILARITY_CREDENTIAL2AUDIO.value
        audio = self._get_virtual_audio_file(data_model.audio_to_evaluate)
        files = {
            "audio_to_evaluate": ("audio", audio, "audio/wav"),
        }
        data = {
            "credential_reference": data_model.credential_reference,
            "channel": data_model.channel,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data, files=files)
        return CompareCredential2AudioOutput(status_code=response.status_code, **response.json())

    def _compare_audio2audio(self, data_model: CompareAudio2AudioInput) -> CompareAudio2AudioOutput:
        """Compare two audio files.

        Args:
            data_model: The data required to compare the audio files

        Returns:
            CompareAudio2AudioOutput: The response from the service

        """
        endpoint = DaspeakEndpoints.SIMILARITY_AUDIO2AUDIO.value
        audio_reference = self._get_virtual_audio_file(data_model.audio_reference)
        audio_to_evaluate = self._get_virtual_audio_file(data_model.audio_to_evaluate)
        files = {
            "audio_reference": ("audio", audio_reference, "audio/wav"),
            "audio_to_evaluate": ("audio", audio_to_evaluate, "audio/wav"),
        }
        data = {
            "channel_reference": data_model.channel_reference,
            "channel_to_evaluate": data_model.channel_to_evaluate,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data, files=files)
        return CompareAudio2AudioOutput(status_code=response.status_code, **response.json())

    def _compare_credential2credential(
            self,
            data_model: CompareCredential2CredentialInput,
        ) -> CompareCredential2CredentialOutput:
        """Compare two credentials.

        Args:
            data_model: The data required to compare the credentials

        Returns:
            CompareCredential2CredentialOutput: The response from the service

        """
        endpoint = DaspeakEndpoints.SIMILARITY_CREDENTIAL2CREDENTIAL.value
        data = {
            "credential_reference": data_model.credential_reference,
            "credential_to_evaluate": data_model.credential_to_evaluate,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data)
        return CompareCredential2CredentialOutput(status_code=response.status_code, **response.json())

    def _compare_audio2credentials(
        self,
        data_model: CompareAudio2CredentialsInput,
    ) -> CompareAudio2CredentialsOutput:
        """Compare an audio file with a list of credentials.

        Args:
            data_model: The data required to compare the audio file with the credentials

        Returns:
            CompareAudio2CredentialsOutput: The response from the service

        """
        endpoint = DaspeakEndpoints.IDENTIFICATION_AUDIO2CREDENTIALS.value
        audio = self._get_virtual_audio_file(data_model.audio_reference)
        files = {
            "audio_reference": ("audio_reference", audio, "audio/wav"),
        }
        credential_list = json.dumps(data_model.credential_list)
        data = {
            "credential_list": credential_list,
            "channel": data_model.channel,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data, files=files)
        return CompareAudio2CredentialsOutput(status_code=response.status_code, **response.json())

    def _compare_credential2credentials(
        self,
        data_model: CompareCredential2CredentialsInput,
    ) -> CompareCredential2CredentialsOutput:
        endpoint = DaspeakEndpoints.IDENTIFICATION_CREDENTIAL2CREDENTIALS.value
        credential_list = json.dumps(data_model.credential_list)
        data = {
            "credential_reference": data_model.credential_reference,
            "credential_list": credential_list,
            "calibration": data_model.calibration,
        }
        response = self._post(endpoint=endpoint, data=data)
        return CompareCredential2CredentialsOutput(status_code=response.status_code, **response.json())

    def _get_virtual_audio_file(self, audio_input: object) -> bytes:
        if isinstance(audio_input, str):
            try:
                with open(audio_input, "rb") as f:
                    audio = f.read()
            except FileNotFoundError as e:
                error = f"File {audio_input} not found"
                raise FileNotFoundError(error) from e
        elif isinstance(audio_input, bytes):
            audio = audio_input
        else:
            error = "audio must be a string or a bytes object"
            raise TypeError(error)
        return audio
